{
 DX8PartSys ---- Pulsar DirectX 8 Weird And Simple Particle System :P
                 by Lord Trancos.
}

unit DX8PartSys;

// --------------------------------------------------------------------

interface

uses Windows, DirectXGraphics, D3DX8, DX8GFX, MathVect;

const
  D3DFVF_PSVERTEX = D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1;

type
  // Emitter type; Point or Axis Aligned Box
  TPSEmitterType = (etPoint, etAAB);

  // Axis Alignment
  TPSAxisAlignment = (aaNone, aaY);

  // Emitter
  PPSEmitter = ^TPSEmitter;
  TPSEmitter = record
                 Typ : TPSEmitterType;
                 Min, Max : TD3DVector; // For AAB Emitter
                 Inc : TD3DVector;      // For AAB Emitter
                 Org : TD3DVector;      // For Point Emitter
               end;

  // Particle
  PPSParticle = ^TPSParticle;
  TPSParticle = record
                  Pos0, Vel0 : TD3DVector; // Initial values
                  Color0     : dWord;      // Initial color value
                  Pos        : TD3DVector; // Particle position
                  Scale      : TD3DVector; // Particle scale
                  D          : single;     // Used for sorting
                  Offset     : cardinal;   // Quad offset inside vertex buffer
                  Age        : cardinal;
                end;

  PPSAParticle = ^TPSAParticle;
  TPSAParticle = array[0..$FFFF-1] of TPSParticle;

  // Particle System
  PPSSystem = ^TPSSystem;
  TPSSystem = record
                Emitter    : TPSEmitter;       // Emitter
                Alignment  : TPSAxisAlignment; // Billboard alignment
                nParticles : cardinal;         // # of particles
                Particles  : pointer;          // particle buffer
                PartVB     : IDirect3DVertexBuffer8; // Vertex Buffer
                Billboard  : TD3DMatrix;       // Billboard matrix
                Age        : cardinal;         // Particle system age
                Ok         : boolean;
              end;

  // Particle Vertex
  TPSVertex = record
                pos   : TD3DVector;
                color : dWord;
                u, v  : single;
              end;

  // Particle Quad
  PPSQuad = ^TPSQuad;
  TPSQuad = array[0..3] of TPSVertex;

  TPSNewPartCB = procedure(const _p: PPSParticle); stdcall;
  TPSAniPartCB = procedure(const _p: PPSParticle;
                           const _q: PPSQuad); stdcall;

function  psEmitter_Point(_p: TD3DVector): TPSEmitter;
function  psEmitter_AAB(_min, _max: TD3DVector): TPSEmitter;

function  psNewParticlePos(_emit: PPSEmitter): TD3DVector;
procedure psNewParticle(_emit : PPSEmitter;
                        _part : PPSParticle;
                        _partColor : dWord;
                        _newPartCB : TPSNewPartCB);

procedure psResetSystem(var _ps: TPSSystem);
procedure psFreeSystem(var _ps: TPSSystem);
function  psNewSystem(_emit  : TPSEmitter;
                      _align : TPSAxisAlignment;
                      _numParticles : word;
                      _defPartColor : dWord;
                      _defPartWidth, _defPartHeight: single;
                      _newPartCB: TPSNewPartCB): TPSSystem;
function  psUpdateSystem(var _ps : TPSSystem; _time: dWord;
                         _lockVB : boolean;
                         _cameraPos, _cameraTarget: TD3DVector;
                         _aniCB : TPSAniPartCB): boolean;
function  psHiddenAnimation(var _ps : TPSSystem;
                            _nSteps, _stepInterval: dWord;
                            _lockVB : boolean;
                            _aniCB : TPSAniPartCB): boolean;
procedure psRenderSystem(_ps : TPSSystem);

// --------------------------------------------------------------------

implementation

// --------------------------------------------------------------------

function psEmitter_Point(_p: TD3DVector): TPSEmitter;

begin
  Result.Typ := etPoint;
  Result.Org := _p;
end;

// --------------------------------------------------------------------

function psEmitter_AAB(_min, _max: TD3DVector): TPSEmitter;

begin
  Result.Typ := etAAB;
  Result.Min := _min;
  Result.Max := _max;
  Result.Inc := vectSubV(_max, _min);
end;

// --------------------------------------------------------------------

function psNewParticlePos(_emit: PPSEmitter): TD3DVector;


begin
  Case _emit^.Typ of
    etPoint  : Result := _emit^.Org;
    etAAB    : begin
                 Result.x := _emit^.Min.x + (_emit^.Inc.x * Random);
                 Result.y := _emit^.Min.y + (_emit^.Inc.y * Random);
                 Result.z := _emit^.Min.z + (_emit^.Inc.z * Random);
               end;
  end;
end;

// --------------------------------------------------------------------

procedure psNewParticle(_emit : PPSEmitter;
                        _part : PPSParticle;
                        _partColor : dWord;
                        _newPartCB : TPSNewPartCB);

begin
  _part^.Pos0   := psNewParticlePos(_emit);
  _part^.Vel0   := vect(0, 0, 0);
  _part^.Color0 := _partColor;
  _part^.Pos    := _part^.Pos0;
  _part^.Scale  := vect(1, 1, 1);
  _part^.Age    := 0;
  if assigned(_newPartCB) then _newPartCB(_part);
end;

// --------------------------------------------------------------------

procedure psResetSystem(var _ps: TPSSystem);

begin
  FillChar(_ps, SizeOf(_ps), 0);
end;

// --------------------------------------------------------------------

procedure psFreeSystem(var _ps: TPSSystem);

begin
  // Free
  if (_ps.Particles <> NIL)
    then FreeMem(_ps.Particles, SizeOf(TPSParticle) * _ps.nParticles);
  _ps.PartVB := NIL;
  // Reset
  psResetSystem(_ps);
end;

// --------------------------------------------------------------------

function psNewSystem(_emit  : TPSEmitter;
                     _align : TPSAxisAlignment;
                     _numParticles : word;
                     _defPartColor : dWord;
                     _defPartWidth, _defPartHeight: single;
                     _newPartCB: TPSNewPartCB): TPSSystem;

var _cnt : cardinal;
    _p : PPSParticle;
    _q : PPSQuad;
    _i : cardinal;

begin
  psResetSystem(Result);

  // Set Emitter
  Result.Emitter    := _emit;
  Result.Alignment  := _align;
  Result.nParticles := _numParticles;
  Result.Age        := 0;

  // Create buffer
  try
    GetMem(Result.Particles, SizeOf(TPSParticle) * _numParticles);
  except
    psFreeSystem(Result);
    exit;
  end;

  // Create particles
  _p := Result.Particles;
  for _cnt := 0 to _numParticles-1 do
  begin
    psNewParticle(@_emit, _p, _defPartColor, _newPartCB);
    _p^.Offset := (_cnt*4);
    // Next particle
    inc(_p);
  end;

  // Create and Lock VB
  if (failed(D3DDEV8.CreateVertexBuffer(_numParticles*4*SizeOf(TPSVertex),
                                        D3DUSAGE_WRITEONLY, D3DFVF_PSVERTEX,
                                        D3DPOOL_DEFAULT, Result.PartVB))) or
     (failed(Result.PartVB.Lock(0, 0, pByte(_q), 0))) then
  begin
    psFreeSystem(Result);
    exit;
  end;

  // Fill
  _p := Result.Particles;
  for _i := 0 to _numParticles-1 do
  begin
    _q^[0].pos   := vect(-_defPartWidth, -_defPartHeight, 0.0);
    _q^[0].color := _p^.Color0;
    _q^[0].u     := 0.0;
    _q^[0].v     := 1.0;
    _q^[1].pos   := vect(-_defPartWidth, +_defPartHeight, 0.0);
    _q^[1].color := _p^.Color0;
    _q^[1].u     := 0.0;
    _q^[1].v     := 0.0;
    _q^[2].pos   := vect(+_defPartWidth, -_defPartHeight, 0.0);
    _q^[2].color := _p^.Color0;
    _q^[2].u     := 1.0;
    _q^[2].v     := 1.0;
    _q^[3].pos   := vect(+_defPartWidth, +_defPartHeight, 0.0);
    _q^[3].color := _p^.Color0;
    _q^[3].u     := 1.0;
    _q^[3].v     := 0.0;

    // Next particle/quad
    inc(_p);
    inc(_q);
  end;

  // Unlock
  Result.PartVB.Unlock();

  // All right
  Result.Ok := true;
end;

// --------------------------------------------------------------------

function psUpdateSystem(var _ps : TPSSystem; _time: dWord;
                        _lockVB : boolean;
                        _cameraPos, _cameraTarget: TD3DVector;
                        _aniCB : TPSAniPartCB): boolean;

var _dir : TD3DVector;
    _ap : PPSAParticle;
    _p : PPSParticle;
    _q, _r : PPSQuad;
    _i : cardinal;

  // ---------------------------------------

  function GetBillboardMatrix: TD3DMatrix;

  // (For non axis aligned billboarding)

  var _angle : single;
      _worldUp, _right, _up, _view : TD3DVector;

  begin
    _view    := _dir;
    _worldUp := vect(0, 1, 0);
    _angle   := vectDot(_worldUp, _view);
    _up      := vectNormalize(vectSubV(_worldUp, vectMulS(_view, _angle)));
    _right   := vectCross(_up, _view);

    with Result do
    begin
      _11 := _right.x; _21 := _up.x; _31 := _view.x;
      _12 := _right.y; _22 := _up.y; _32 := _view.y;
      _13 := _right.z; _23 := _up.z; _33 := _view.z;
      _14 := 0.0; _24 := 0.0; _34 := 0.0; _44 := 1.0;
    end;
  end;

  // ---------------------------------------

  procedure QuickSort(_iLo, _iHi: longint);

  var _lo, _hi : longint;
      _mid : single;
      _t : TPSParticle;

  begin
    _lo  := _iLo;
    _hi  := _iHi;
    _mid := _ap^[(_lo + _hi) div 2].D;
    repeat
      while _ap^[_lo].d < _mid do inc(_lo);
      while _ap^[_hi].d > _mid do dec(_hi);
      if _lo <= _hi then
      begin
        // Swap
        _t := _ap^[_lo];
        _ap^[_lo] := _ap^[_hi];
        _ap^[_hi] := _t;
        inc(_lo);
        dec(_hi);
      end;
    until _lo > _hi;
    if _hi > _iLo then QuickSort(_iLo, _hi);
    if _lo < _iHi then QuickSort(_Lo, _iHi);
  end;

  // ---------------------------------------

begin
  Result := false;

  // Lock vertex buffer if needed
  _q := NIL;
  if (_lockVB) then
    if (failed(_ps.PartVB.Lock(0, 0, pByte(_q), 0))) then exit;

  // Increase system age
  inc(_ps.Age, _time);

  // Animate particles
  _p := _ps.Particles;
  for _i := 0 to _ps.nParticles-1 do
  begin
    inc(_p^.Age, _time);
    if Assigned(_aniCB) then
    begin
      _r := _q;
      if (_r <> NIL) then inc(_r, _p^.offset shr 2);
      _aniCB(_p, _r);
    end;
    // next
    inc(_p);
  end;

  // Unlock VB
  if (_lockVB) then _ps.PartVB.Unlock;

  // Get billboard direction
  _dir := vectNormalize(vectSubV(_cameraTarget, _cameraPos));

  // --
  _p  := _ps.Particles;
  _ap := _ps.Particles;

  // Y axis aligned
  if (_ps.Alignment = aaY) then
  begin
    // billboard
    if (_dir.x > 0)
      then D3DXMatrixRotationY(_ps.Billboard, -arctan(_dir.z/_dir.x)+HALF_PI)
        else D3DXMatrixRotationY(_ps.Billboard, -arctan(_dir.z/_dir.x)-HALF_PI);
    // d values
    for _i := 0 to _ps.nParticles-1 do
    begin
      _p^.D := _p^.Pos.x * _dir.x + _p^.Pos.z * _dir.z;
      inc(_p); // next
    end;
  end else
  // Non axis aligned
  begin
    // billboard
    _ps.Billboard := GetBillboardMatrix;
    // d values
    for _i := 0 to _ps.nParticles-1 do
    begin
      _p^.D := _p^.Pos.x * _dir.x + _p^.Pos.y * _dir.y + _p^.Pos.z * _dir.z;
      inc(_p); // next
    end;
  end;

  // Sort
  QuickSort(0, _ps.nParticles-1);

  // All right
  Result := true;
end;

// --------------------------------------------------------------------

function psHiddenAnimation(var _ps : TPSSystem;
                           _nSteps, _stepInterval: dWord;
                           _lockVB : boolean;
                           _aniCB : TPSAniPartCB): boolean;

// Used for particle system initializing - (does a hidden animation)

var
  _p : PPSParticle;
  _q, _r : PPSQuad;
  _i, _j : cardinal;

begin
  Result := false;

  // Lock vertex buffer if needed
  _q := NIL;
  if (_lockVB) then
    if (failed(_ps.PartVB.Lock(0, 0, pByte(_q), 0))) then exit;

  // ---
  for _j := 0 to _nSteps - 1 do
  begin
    // Increase system age
    inc(_ps.Age, _stepInterval);

    // Animate particles
    _p := _ps.Particles;
    for _i := 0 to _ps.nParticles-1 do
    begin
      inc(_p^.Age, _stepInterval);
      if Assigned(_aniCB) then
      begin
        _r := _q;
        if (_r <> NIL) then inc(_r, _p^.offset shr 2);
        _aniCB(_p, _r);
      end;
      // next
      inc(_p);
    end;
  end;

  // Unlock VB
  if (_lockVB) then _ps.PartVB.Unlock;

  // All right
  Result := true;
end;

// --------------------------------------------------------------------

procedure psRenderSystem(_ps : TPSSystem);

var _cnt : cardinal;
    _mtx, _final : TD3DMatrix;
    _p : PPSParticle;

begin
  if not _ps.ok then exit;

  // Set vertex shader and stream source
  D3DDEV8.SetVertexShader(D3DFVF_PSVERTEX);
  D3DDEV8.SetStreamSource(0, _ps.PartVB, sizeof(TPSVertex));

  // Loop through and render all particles
  _p := _ps.Particles;
  inc(_p, _ps.nParticles-1);
  for _cnt := 0 to _ps.nParticles-1 do
  begin
    // Translate the billboard into place
    _ps.Billboard._41 := _p^.pos.x;
    _ps.Billboard._42 := _p^.pos.y;
    _ps.Billboard._43 := _p^.pos.z;
    // Scale
    D3DXMatrixScaling(_mtx, _p^.Scale.x, _p^.Scale.y, _p^.Scale.z);
    D3DXMatrixMultiply(_final, _mtx, _ps.Billboard);
    // Set billboard matrix
    D3DDEV8.SetTransform(D3DTS_WORLD, _final);
    // Render the billboard
    D3DDEV8.DrawPrimitive(D3DPT_TRIANGLESTRIP, _p^.Offset, 2);
    // Next
    dec(_p);
  end;
end;

// --------------------------------------------------------------------

begin
  Randomize; // ;) All particle engine needs it
end.
